import Orange

import numpy as np
np.set_printoptions(linewidth = 150)

print("Load, normalize and split data.")

data = Orange.data.Table("Ames.tab")
normalizer = Orange.preprocess.Normalize()
data = normalizer(data)
n = len(data)
print("{} Instances, {} Features".format(n, len(data.domain.attributes)))
""" 6494 Instances, 165 Features """


np.random.seed(0)
split = np.random.choice([True, False], n, p = [0.8, 0.2])
train, test = data[split], data[~split]
print("Train size: {}, Test size: {}".format(len(train), len(test)))
""" Train size: 5220, Test size: 1274 """


print("\nEvaluate different classification methods.")
learners = [
   Orange.classification.LogisticRegressionLearner(),
   Orange.classification.NaiveBayesLearner(),
   Orange.classification.KNNLearner(),
   Orange.classification.RandomForestLearner(),
   Orange.classification.SVMLearner(),
]
res = Orange.evaluation.TestOnTestData(train, test, learners)
ev = np.stack((list(map(str, learners)),
              Orange.evaluation.CA(res),
              Orange.evaluation.AUC(res),
              Orange.evaluation.Precision(res),
              Orange.evaluation.Recall(res),
              Orange.evaluation.F1(res)), axis = -1)
print(ev)
"""
[['logistic regression' '0.7480376766091051' '0.8189954518979342' '0.7732732732732732' '0.7518248175182481' '0.7623982235381199']
 ['naive bayes' '0.6687598116169545' '0.7212248893956105' '0.7057902973395931' '0.6583941605839416' '0.6812688821752266']
 ['knn' '0.7747252747252747' '0.8434709330425192' '0.7875722543352601' '0.7956204379562044' '0.7915758896151054']
 ['random forest' '0.7653061224489796' '0.8391173955609531' '0.8133116883116883' '0.7313868613138687' '0.7701767870868563']
 ['svm' '0.7723704866562009' '0.7724982340475629' '0.7987897125567323' '0.7708029197080292' '0.7845468053491826']]
"""


print("\nRunning conformal prediction.")
import orangecontrib.conformal as cp

nc = cp.nonconformity.InverseProbability(Orange.classification.LogisticRegressionLearner())
ic = cp.classification.InductiveClassifier(nc)
r = None

import pickle
if True:
    r = cp.evaluation.run_train_test(ic, 0.1, train[0::2], test, train[1::2])
    with open('r.pickle', 'wb') as f:
        pickle.dump(r, f)
else:
    with open('r.pickle', 'rb') as f:
        r = pickle.load(f)
print("Accuracy at eps=0.1:", r.accuracy())


print("\nDraw validation plot.")
import matplotlib.pyplot as plt
plt.figure(figsize = (4, 3))
plt.xlabel('Significance level (eps)')
plt.ylabel('Accuracy')
x = np.linspace(0.0, 1.0, 11)
y = [round(r.accuracy(eps = e), 3) for e in x]
plt.plot(x, y, linestyle = '-', marker = 'o', color = 'b')
plt.tight_layout()
plt.savefig("Figures/validation.eps")


print("\nEvaluate different conformal prediction methods and nonconformities.")

import warnings
# LogisticRegressionLearner that can handle data with a single class value
class LogReg(Orange.classification.LogisticRegressionLearner):
      def __call__(self, data):
            if len(np.unique(data.Y)) > 1:
                  return super().__call__(data)
            else:
                  warnings.warn("Single class in data, returning Constant Model.")
                  maj = Orange.classification.MajorityLearner()
                  const = maj(data)
                  return const

import sklearn
ncs = [
    cp.nonconformity.InverseProbability(LogReg()),
    cp.nonconformity.ProbabilityMargin(LogReg()),
    cp.nonconformity.SVMDistance(sklearn.svm.SVC()),
    cp.nonconformity.KNNDistance(k = 10),
    cp.nonconformity.KNNFraction(k = 10, weighted = True),
    cp.nonconformity.LOOClassNC(LogReg(), k = 10)
]

train, cal = train[0::2], train[1::2]
for eps in [0.05, 0.1, 0.2]:
    for nc in ncs:
        ic = cp.classification.InductiveClassifier(nc)
        r = cp.evaluation.run_train_test(ic, eps, train, test, calibrate = cal)
        print(eps, nc, round(r.accuracy(), 3), round(r.singleton_criterion(), 3),
              round(r.singleton_correct(),3), round(r.multiple_criterion(), 3),
              round(r.empty_criterion(), 3), round(r.confidence(), 3),
              round(r.credibility(), 3))
"""
0.05 InverseProbability (logistic regression) 0.941 0.405 0.855 0.595 0.0 0.91 0.58
0.05 ProbabilityMargin (logistic regression) 0.941 0.405 0.855 0.595 0.0 0.91 0.58
0.05 SVMDistance (SVC) 0.947 0.458 0.883 0.542 0.0 0.918 0.572
0.05 KNNDistance (k=10) 0.96 0.389 0.897 0.611 0.0 0.91 0.569
0.05 KNNFraction (k=10) 0.951 0.452 0.891 0.548 0.0 0.911 0.591
0.05 LOOClassNC 0.942 0.44 0.877 0.556 0.004 0.907 0.583

0.1 InverseProbability (logistic regression) 0.889 0.614 0.82 0.386 0.0 0.91 0.58
0.1 ProbabilityMargin (logistic regression) 0.889 0.614 0.82 0.386 0.0 0.91 0.58
0.1 SVMDistance (SVC) 0.903 0.641 0.849 0.359 0.0 0.918 0.572
0.1 KNNDistance (k=10) 0.908 0.611 0.85 0.389 0.0 0.91 0.569
0.1 KNNFraction (k=10) 0.904 0.627 0.847 0.373 0.0 0.911 0.591
0.1 LOOClassNC 0.905 0.597 0.851 0.397 0.006 0.907 0.583

0.2 InverseProbability (logistic regression) 0.792 0.884 0.765 0.116 0.0 0.91 0.58
0.2 ProbabilityMargin (logistic regression) 0.792 0.884 0.765 0.116 0.0 0.91 0.58
0.2 SVMDistance (SVC) 0.794 0.922 0.777 0.078 0.0 0.918 0.572
0.2 KNNDistance (k=10) 0.808 0.883 0.783 0.117 0.0 0.91 0.569
0.2 KNNFraction (k=10) 0.81 0.874 0.783 0.126 0.0 0.911 0.591
0.2 LOOClassNC 0.806 0.832 0.812 0.13 0.038 0.907 0.583
"""
